@@ -1,6 +1,6 @@
 // 
 //
-//  $Id: spamass-milter.cpp,v 1.52 2003/06/26 15:10:44 dnelson Exp $
+//  $Id: spamass-milter.cpp,v 1.86 2005/02/05 07:03:22 dnelson Exp $
 //
 //  SpamAss-Milter 
 //    - a rather trivial SpamAssassin Sendmail Milter plugin
@@ -75,26 +75,19 @@
 #include <stdlib.h>
 #include <stdarg.h>
 #include <string.h>
+#include <strings.h>
 #include <sysexits.h>
 #include <unistd.h>
 #include <fcntl.h>
 #include <syslog.h>
 #include <signal.h>
+#include <pthread.h>
 #ifdef HAVE_POLL_H
 #include <poll.h>
 #else
 #include "subst_poll.h"
 #endif
 #include <errno.h>
-#ifndef HAVE_STRSEP
-#ifdef  __cplusplus
-extern "C" {
-#endif
-char *strsep(char **, const char *);
-#ifdef  __cplusplus
-}
-#endif
-#endif
 
 // C++ includes
 #include <cstdio>
@@ -114,6 +107,10 @@ extern "C" {
 char *strsep(char **stringp, const char *delim);
 #endif 
 
+#if !HAVE_DECL_DAEMON
+int daemon(int nochdir, int noclose);
+#endif 
+
 #ifdef  __cplusplus
 }
 #endif
@@ -130,7 +127,7 @@ char *strsep(char **stringp, const char *delim);
 
 // }}} 
 
-static const char Id[] = "$Id: spamass-milter.cpp,v 1.52 2003/06/26 15:10:44 dnelson Exp $";
+static const char Id[] = "$Id: spamass-milter.cpp,v 1.86 2005/02/05 07:03:22 dnelson Exp $";
 
 struct smfiDesc smfilter =
   {
@@ -150,16 +147,19 @@ struct smfiDesc smfilter =
   };
 
 const char *const debugstrings[] = {
-	"ALL", "FUNC", "POLL", "UORI", "STR", "MISC", "NET", "SPAMC",
+	"ALL", "FUNC", "POLL", "UORI", "STR", "MISC", "NET", "SPAMC", "RCPT",
+	"COPY",
 	NULL
 };
 
 int flag_debug = (1<<D_ALWAYS);
 bool flag_reject = false;
 int reject_score = -1;
-bool dontmodify = false;
+bool dontmodifyspam = false;    // Don't modify/add body or spam results headers
+bool dontmodify = false;        // Don't add SA headers, ever.
 bool flag_sniffuser = false;
-char *defaultuser;
+char *defaultuser;				/* Username to send to spamc if there are multiple recipients */
+char *defaultdomain;			/* Domain to append if incoming address has none */
 char *spamdhost;
 struct networklist ignorenets;
 int spamc_argc;
@@ -167,6 +167,12 @@ char **spamc_argv;
 bool flag_bucket = false;
 bool flag_bucket_only = false;
 char *spambucket;
+bool flag_full_email = false;		/* pass full email address to spamc */
+bool flag_expand = false;	/* alias/virtusertable expansion */
+
+#if defined(__FreeBSD__) /* popen bug - see PR bin/50770 */
+static pthread_mutex_t popen_mutex = PTHREAD_MUTEX_INITIALIZER;
+#endif
 
 // {{{ main()
 
@@ -174,18 +180,21 @@ int
 main(int argc, char* argv[])
 {
    int c, err = 0;
-   const char *args = "p:fd:mr:u:D:i:b:B:";
+   const char *args = "fd:mMp:P:r:u:D:i:b:B:e:x";
    char *sock = NULL;
    bool dofork = false;
+   char *pidfilename = NULL;
+   FILE *pidfile = NULL;
+
+#ifdef HAVE_VERBOSE_TERMINATE_HANDLER
+	std::set_terminate (__gnu_cxx::__verbose_terminate_handler);
+#endif
 
    openlog("spamass-milter", LOG_PID, LOG_MAIL);
 
 	/* Process command line options */
 	while ((c = getopt(argc, argv, args)) != -1) {
 		switch (c) {
-			case 'p':
-				sock = strdup(optarg);
-				break;
 			case 'f':
 				dofork = true;
 				break;
@@ -195,13 +204,28 @@ main(int argc, char* argv[])
 			case 'D':
 				spamdhost = strdup(optarg);
 				break;
+			case 'e':
+				flag_full_email = true;
+				defaultdomain = strdup(optarg);
+				break;
 			case 'i':
 				debug(D_MISC, "Parsing ignore list");
 				parse_networklist(optarg, &ignorenets);
 				break;
 			case 'm':
+				dontmodifyspam = true;
+				smfilter.xxfi_flags &= ~SMFIF_CHGBODY;
+				break;
+			case 'M':
 				dontmodify = true;
-				// smfilter.xxfi_flags &= ~SMFIF_CHGBODY;
+				dontmodifyspam = true;
+				smfilter.xxfi_flags &= ~(SMFIF_CHGBODY|SMFIF_CHGHDRS);
+				break;
+			case 'p':
+				sock = strdup(optarg);
+				break;
+			case 'P':
+				pidfilename = strdup(optarg);
 				break;
 			case 'r':
 				flag_reject = true;
@@ -232,12 +256,21 @@ main(int argc, char* argv[])
 				// XXX we should probably verify that optarg is vaguely sane
 				spambucket = strdup( optarg );
 				break;
+			case 'x':
+				flag_expand = true;
+				break;
 			case '?':
 				err = 1;
 				break;
 		}
 	}
 
+   if (flag_full_email && !flag_sniffuser)
+   {
+   	  fprintf(stderr, "-e flag requires -u\n");
+      err=1;
+   }
+
    /* remember the remainer of the arguments so we can pass them to spamc */
    spamc_argc = argc - optind;
    spamc_argv = argv + optind;
@@ -245,39 +278,65 @@ main(int argc, char* argv[])
    if (!sock || err) {
       cout << PACKAGE_NAME << " - Version " << PACKAGE_VERSION << endl;
       cout << "SpamAssassin Sendmail Milter Plugin" << endl;
-      cout << "Usage: spamass-milter -p socket [-d nn] [-D host] [-f] [-i networks] [-m]" << endl;
-      cout << "                      [-r nn] [-u defaultuser] [-b|-B bucket ] [-- spamc args ]" << endl;
+      cout << "Usage: spamass-milter -p socket [-b|-B bucket] [-d xx[,yy...]] [-D host]" << endl;
+      cout << "                      [-e defaultdomain] [-f] [-i networks] [-m] [-M]" << endl;
+      cout << "                      [-P pidfile] [-r nn] [-u defaultuser] [-x]" << endl;
+      cout << "                      [-- spamc args ]" << endl;
       cout << "   -p socket: path to create socket" << endl;
       cout << "   -b bucket: redirect spam to this mail address.  The orignal" << endl;
       cout << "          recipient(s) will not receive anything." << endl;
       cout << "   -B bucket: add this mail address as a BCC recipient of spam." << endl;
       cout << "   -d xx[,yy ...]: set debug flags.  Logs to syslog" << endl;
-      cout << "   -D host: connect to spand at remote host (deprecated)" << endl;
+      cout << "   -D host: connect to spamd at remote host (deprecated)" << endl;
+      cout << "   -e defaultdomain: pass full email address to spamc instead of just\n"
+              "          username.  Uses 'defaultdomain' if there was none" << endl;
       cout << "   -f: fork into background" << endl;
       cout << "   -i: skip (ignore) checks from these IPs or netblocks" << endl;
-      cout << "       example: -i 192.168.12.5,10.0.0.0/8,172.16/255.255.0.0" << endl;
+      cout << "          example: -i 192.168.12.5,10.0.0.0/8,172.16/255.255.0.0" << endl;
       cout << "   -m: don't modify body, Content-type: or Subject:" << endl;
+      cout << "   -M: don't modify the message at all" << endl;
+      cout << "   -P pidfile: Put processid in pidfile" << endl;
       cout << "   -r nn: reject messages with a score >= nn with an SMTP error.\n"
               "          use -1 to reject any messages tagged by SA." << endl;
       cout << "   -u defaultuser: pass the recipient's username to spamc.\n"
               "          Uses 'defaultuser' if there are multiple recipients." << endl;
+      cout << "   -x: pass email address through alias and virtusertable expansion." << endl;
       cout << "   -- spamc args: pass the remaining flags to spamc." << endl;
               
       exit(EX_USAGE);
    }
 
-	if (dofork == true) {
-		switch(fork()) {
-         case -1: /* Uh-oh, we have a problem forking. */
-            fprintf(stderr, "Uh-oh, couldn't fork!\n");
-				exit(errno);
-				break;
-			case 0: /* Child */
-				break;
-			default: /* Parent */
-				exit(0);
+	if (pidfilename)
+	{
+		unlink(pidfilename);
+		pidfile = fopen(pidfilename,"w");
+		if (!pidfile)
+		{
+			fprintf(stderr, "Could not open pidfile: %s\n", strerror(errno));
+			exit(1);
+		}
+		/* leave the file open through the fork, since we don't know our pid
+		   yet
+		*/
+	}
+
+
+	if (dofork == true) 
+	{
+		if (daemon(0, 0) == -1)
+		{
+            fprintf(stderr, "daemon() failed: %s\n", strerror(errno));
+            exit(1);
 		}
 	}
+	
+	if (pidfile)
+	{
+		fprintf(pidfile, "%ld\n", (long)getpid());
+		fclose(pidfile);
+		pidfile = NULL;
+	}	
+	
    {
       struct stat junk;
       if (stat(sock,&junk) == 0) unlink(sock);
@@ -293,6 +352,8 @@ main(int argc, char* argv[])
 	debug(D_ALWAYS, "spamass-milter %s starting", PACKAGE_VERSION);
 	err = smfi_main();
 	debug(D_ALWAYS, "spamass-milter %s exiting", PACKAGE_VERSION);
+	if (pidfilename)
+		unlink(pidfilename);
 	return err;
 }
 
@@ -301,8 +362,8 @@ main(int argc, char* argv[])
 /* Update a header if SA changes it, or add it if it is new. */
 void update_or_insert(SpamAssassin* assassin, SMFICTX* ctx, string oldstring, t_setter setter, char *header )
 {
-	string::size_type eoh1(assassin->d().find("\n\n"));
-	string::size_type eoh2(assassin->d().find("\n\r\n"));
+	string::size_type eoh1 = assassin->d().find("\n\n");
+	string::size_type eoh2 = assassin->d().find("\n\r\n");
 	string::size_type eoh = ( eoh1 < eoh2 ? eoh1 : eoh2 );
 
 	string newstring;
@@ -311,28 +372,31 @@ void update_or_insert(SpamAssassin* assassin, SMFICTX* ctx, string oldstring, t_
 	debug(D_UORI, "u_or_i: looking at <%s>", header);
 	debug(D_UORI, "u_or_i: oldstring: <%s>", oldstring.c_str());
 
-	newstring = retrieve_field(assassin->d().substr(0, eoh), string(header));
+	newstring = retrieve_field(assassin->d().substr(0, eoh), header);
 	debug(D_UORI, "u_or_i: newstring: <%s>", newstring.c_str());
 
 	oldsize = callsetter(*assassin,setter)(newstring);
       
-	if (newstring != oldstring)
+	if (!dontmodify)
 	{
-		/* change if old one was present, append if non-null */
-		if (oldsize > 0)
+		if (newstring != oldstring)
 		{
-			debug(D_UORI, "u_or_i: changing");
-			smfi_chgheader(ctx, header, 1, newstring.size() > 0 ? 
-				const_cast<char*>(newstring.c_str()) : NULL );
-		} else if (newstring.size() > 0)
+			/* change if old one was present, append if non-null */
+			char* cstr = const_cast<char*>(newstring.c_str());
+			if (oldsize > 0)
+			{
+				debug(D_UORI, "u_or_i: changing");
+				smfi_chgheader(ctx, header, 1, newstring.size() > 0 ? 
+					cstr : NULL );
+			} else if (newstring.size() > 0)
+			{
+				debug(D_UORI, "u_or_i: inserting");
+				smfi_addheader(ctx, header, cstr);
+			}
+		} else
 		{
-			debug(D_UORI, "u_or_i: inserting");
-			smfi_addheader(ctx, header, 
-				const_cast<char*>(newstring.c_str()));
+			debug(D_UORI, "u_or_i: no change");
 		}
-	} else
-	{
-		debug(D_UORI, "u_or_i: no change");
 	}
 }
 
@@ -346,11 +410,14 @@ assassinate(SMFICTX* ctx, SpamAssassin* assassin)
 {
   // find end of header (eol in last line of header)
   // and beginning of body
-  string::size_type eoh1(assassin->d().find("\n\n"));
-  string::size_type eoh2(assassin->d().find("\n\r\n"));
-  string::size_type eoh = ( eoh1 < eoh2 ? eoh1 : eoh2 );
+  string::size_type eoh1 = assassin->d().find("\n\n");
+  string::size_type eoh2 = assassin->d().find("\n\r\n");
+  string::size_type eoh = (eoh1 < eoh2) ? eoh1 : eoh2;
   string::size_type bob = assassin->d().find_first_not_of("\r\n", eoh);
 
+  if (bob == string::npos)
+  	bob = assassin->d().size();
+
   update_or_insert(assassin, ctx, assassin->spam_flag(), &SpamAssassin::set_spam_flag, "X-Spam-Flag");
   update_or_insert(assassin, ctx, assassin->spam_status(), &SpamAssassin::set_spam_status, "X-Spam-Status");
 
@@ -359,13 +426,19 @@ assassinate(SMFICTX* ctx, SpamAssassin* assassin)
   if (flag_reject)
   {
 	bool do_reject = false;
-	if (reject_score == -1 && assassin->spam_flag().size()>0)
+	if (reject_score == -1 && !assassin->spam_flag().empty())
 		do_reject = true;
 	if (reject_score != -1)
 	{
 		int score, rv;
 		const char *spam_status = assassin->spam_status().c_str();
-		rv = sscanf(spam_status,"%*s hits=%d", &score);
+		/* SA 3.0 uses the keyword "score" */
+		rv = sscanf(spam_status,"%*s score=%d", &score);
+		if (rv != 1)
+		{
+			/* SA 2.x uses the keyword "hits" */
+			rv = sscanf(spam_status,"%*s hits=%d", &score);
+		}
 		if (rv != 1)
 			debug(D_ALWAYS, "Could not extract score from <%s>", spam_status);
 		else 
@@ -379,13 +452,75 @@ assassinate(SMFICTX* ctx, SpamAssassin* assassin)
 	{
 		debug(D_MISC, "Rejecting");
 		smfi_setreply(ctx, "550", "5.7.1", "Blocked by SpamAssassin");
+
+
+		if (flag_bucket)
+		{
+			/* If we also want a copy of the spam, shell out to sendmail and
+			   send another copy.  The milter API will not let you send the
+			   message AND return a failure code to the sender, so this is
+			   the only way to do it. */
+#if defined(__FreeBSD__)
+			int rv;
+#endif
+			
+#if defined(HAVE_ASPRINTF)
+			char *buf;
+#else
+			char buf[1024];
+#endif
+			char *fmt="%s \"%s\"";
+			FILE *p;
+
+#if defined(HAVE_ASPRINTF)
+			asprintf(&buf, fmt, SENDMAIL, spambucket);
+#else
+#if defined(HAVE_SNPRINTF)
+			snprintf(buf, sizeof(buf)-1, fmt, SENDMAIL, spambucket);
+#else
+			/* XXX possible buffer overflow here */
+			sprintf(buf, fmt, SENDMAIL, spambucket);
+#endif
+#endif
+
+			debug(D_COPY, "calling %s", buf);
+#if defined(__FreeBSD__) /* popen bug - see PR bin/50770 */
+			rv = pthread_mutex_lock(&popen_mutex);
+			if (rv)
+			{
+				debug(D_ALWAYS, "Could not lock popen mutex: %s", strerror(rv));
+				abort();
+			}		
+#endif
+			p = popen(buf, "w");
+			if (!p)
+			{
+				debug(D_COPY, "popen failed(%s).  Will not send a copy to spambucket", strerror(errno));
+			} else
+			{
+				// Send message provided by SpamAssassin
+				fwrite(assassin->d().c_str(), assassin->d().size(), 1, p);
+				pclose(p); p = NULL;
+			}
+#if defined(__FreeBSD__)
+			rv = pthread_mutex_unlock(&popen_mutex);
+			if (rv)
+			{
+				debug(D_ALWAYS, "Could not unlock popen mutex: %s", strerror(rv));
+				abort();
+			}		
+#endif
+#if defined(HAVE_ASPRINTF)
+			free(buf);
+#endif 
+		}
 		return SMFIS_REJECT;
 	}
   }
 
   /* Drop the message into the spam bucket if it's spam */
   if ( flag_bucket ) {
-        if ( assassin->spam_flag().size() > 0 ) {
+        if (!assassin->spam_flag().empty()) {
           // first, add the spambucket address
           if ( smfi_addrcpt( ctx, spambucket ) != MI_SUCCESS ) {
                 throw string( "Failed to add spambucket to recipients" );
@@ -402,7 +537,7 @@ assassinate(SMFICTX* ctx, SpamAssassin* assassin)
                   // It's not 100% important that this succeeds, so we'll just warn on failure rather than bailing out.
                   if ( smfi_delrcpt( ctx, (char *)assassin->recipients.front().c_str()) != MI_SUCCESS ) {
                         // throw_error really just logs a warning as opposed to actually throw()ing
-                        throw_error( "Failed to remove recipient from the envelope" );
+                        debug(D_ALWAYS, "Failed to remove recipient %s from the envelope", assassin->recipients.front().c_str() );
                   }
                   assassin->recipients.pop_front();
                 }
@@ -424,7 +559,7 @@ assassinate(SMFICTX* ctx, SpamAssassin* assassin)
   //  However, only issue the header replacement calls if the content has
   //  actually changed. If SA didn't change subject or content-type, don't
   //  replace here unnecessarily.
-  if (!dontmodify && assassin->spam_flag().size()>0)
+  if (!dontmodifyspam && assassin->spam_flag().size()>0)
     {
 	  update_or_insert(assassin, ctx, assassin->subject(), &SpamAssassin::set_subject, "Subject");
 	  update_or_insert(assassin, ctx, assassin->content_type(), &SpamAssassin::set_content_type, "Content-Type");
@@ -435,7 +570,7 @@ assassinate(SMFICTX* ctx, SpamAssassin* assassin)
       if ( smfi_replacebody(ctx, (unsigned char *)body.c_str(), body_size) == MI_FAILURE )
 	throw string("error. could not replace body.");
       
-    };
+    }
 
   return SMFIS_CONTINUE;
 }
@@ -446,7 +581,7 @@ string
 old_retrieve_field(const string& header, const string& field)
 {
   // look for beginning of content
-  string::size_type pos = find_nocase(header, string("\n")+field+string(": "));
+  string::size_type pos = find_nocase(header, "\n" + field + ": ");
 
   // return empty string if not found
   if (pos == string::npos)
@@ -456,18 +591,18 @@ old_retrieve_field(const string& header, const string& field)
   }
 
   // look for end of field name
-  pos = find_nocase(header, string(" "), pos) + 1;
+  pos = find_nocase(header, " ", pos) + 1;
   
   string::size_type pos2(pos);
 
   // is field empty? 
-  if (pos2 == find_nocase(header, string("\n"), pos2))
+  if (pos2 == find_nocase(header, "\n", pos2))
     return string("");
 
   // look for end of content
   do {
 
-    pos2 = find_nocase(header, string("\n"), pos2+1);
+    pos2 = find_nocase(header, "\n", pos2+1);
 
   }
   while ( pos2 < string::npos &&
@@ -488,7 +623,7 @@ retrieve_field(const string& header, const string& field)
   string::size_type idx = 0;
 
   while( field_start == string::npos ) {
-	idx = find_nocase( header, field + string(":"), idx );
+	idx = find_nocase( header, field + ":", idx );
 
 	// no match
 	if ( idx == string::npos ) {
@@ -539,6 +674,10 @@ retrieve_field(const string& header, const string& field)
 	}
   }
 
+  /* if the header line ends in \r\n, don't return the \r */
+  if (header[field_end-1] == '\r')
+  	field_end--;
+
   //  Maybe remove the whitespace picked up when a header wraps - this
   //  might actually be a requirement
   return header.substr( field_start, field_end - field_start );
@@ -643,7 +782,7 @@ mlfi_envfrom(SMFICTX* ctx, char** envfrom)
   queueid=smfi_getsymval(ctx,"i");
   if (!queueid)
     queueid="unk";
-  assassin->queueid=string(queueid);
+  assassin->queueid = queueid;
 
   debug(D_MISC, "queueid=%s", queueid);
 
@@ -659,18 +798,91 @@ mlfi_envfrom(SMFICTX* ctx, char** envfrom)
 // stores the first recipient in the spamassassin object and
 // stores all addresses and the number thereof (some redundancy)
 //
+
+
 sfsistat
 mlfi_envrcpt(SMFICTX* ctx, char** envrcpt)
 {
 	struct context *sctx = (struct context*)smfi_getpriv(ctx);
 	SpamAssassin* assassin = sctx->assassin;
-	char **rcpt;
+	FILE *p;
+#if defined(__FreeBSD__)
+	int rv;
+#endif
+
+	debug(D_FUNC, "mlfi_envrcpt: enter");
+
+	if (flag_expand)
+	{
+		/* open a pipe to sendmail so we can do address expansion */
+
+		char buf[1024];
+		char *fmt="%s -bv \"%s\" 2>&1";
+
+#if defined(HAVE_SNPRINTF)
+		snprintf(buf, sizeof(buf)-1, fmt, SENDMAIL, envrcpt[0]);
+#else
+		/* XXX possible buffer overflow here */
+		sprintf(buf, fmt, SENDMAIL, envrcpt[0]);
+#endif
+
+		debug(D_RCPT, "calling %s", buf);
+
+#if defined(__FreeBSD__) /* popen bug - see PR bin/50770 */
+		rv = pthread_mutex_lock(&popen_mutex);
+		if (rv)
+		{
+			debug(D_ALWAYS, "Could not lock popen mutex: %s", strerror(rv));
+			abort();
+		}		
+#endif
+
+		p = popen(buf, "r");
+		if (!p)
+		{
+			debug(D_RCPT, "popen failed(%s).  Will not expand aliases", strerror(errno));
+			assassin->expandedrcpt.push_back(envrcpt[0]);
+		} else
+		{
+			while (fgets(buf, sizeof(buf), p) != NULL)
+			{
+				int i = strlen(buf);
+				/* strip trailing EOLs */
+				while (i > 0 && buf[i - 1] <= ' ')
+					i--;
+				buf[i] = '\0';
+				debug(D_RCPT, "sendmail output: %s", buf);
+				/*	From a quick scan of the sendmail source, a valid email
+					address gets printed via either
+					    "deliverable: mailer %s, host %s, user %s"
+					or  "deliverable: mailer %s, user %s"
+				*/
+				if (strstr(buf, "... deliverable: mailer "))
+				{
+					char *p=strstr(buf,", user ");
+					/* anything after ", user " is the email address */
+					debug(D_RCPT, "user: %s", p+7);
+					assassin->expandedrcpt.push_back(p+7);
+				}
+			}
+			pclose(p); p = NULL;
+		}
+#if defined(__FreeBSD__)
+		rv = pthread_mutex_unlock(&popen_mutex);
+		if (rv)
+		{
+			debug(D_ALWAYS, "Could not unlock popen mutex: %s", strerror(rv));
+			abort();
+		}		
+#endif
+	} else
+	{
+		assassin->expandedrcpt.push_back(envrcpt[0]);
+	}	
+	debug(D_RCPT, "Total of %d actual recipients", (int)assassin->expandedrcpt.size());
 
 	if (assassin->numrcpt() == 0)
 	{
-		assassin->set_numrcpt(1);
-		assassin->set_rcpt(string(envrcpt[0]));
-
 		/* Send the envelope headers as X-Envelope-From: and
 		   X-Envelope-To: so that SpamAssassin can use them in its
 		   whitelist checks.  Also forge as complete a dummy
@@ -686,14 +898,13 @@ mlfi_envrcpt(SMFICTX* ctx, char** envrcpt)
 				(envelope-from $g)$.
 		   
 		*/
-		const char *macro_b, *macro_s;
+		const char *macro_b, *macro_s, *macro_j, *macro__;
 
-		/* If the user did not enable the {b} macro in sendmail.cf
-		   just make it blank. Without this date SA can't do
-		   future/past validation on the Date: header */
+		/* Failure to fetch {b} is not fatal.  Without this date SA can't do
+		   future/past validation on the Date: header, but sendmail doesn't
+		   default to allow milters to see it.
+		*/
 		macro_b = smfi_getsymval(ctx, "b");
-		if (!macro_b)
-			macro_b = "";
 			
 		/* Sendmail currently cannot pass us the {s} macro, but
 		   I do not know why.  Leave this in for the day sendmail is
@@ -706,18 +917,43 @@ mlfi_envrcpt(SMFICTX* ctx, char** envrcpt)
 		if (!macro_s)
 			macro_s = "nohelo";
 
+		/* FQDN of this site */
+		macro_j = smfi_getsymval(ctx, "j");
+		if (!macro_j)
+			macro_j = "localhost";
+
+		/* Sending site's address */
+		macro__ = smfi_getsymval(ctx, "_");
+		if (!macro__)
+			macro__ = "unknown";
+
 		assassin->output((string)"X-Envelope-From: "+assassin->from()+"\r\n");
-		assassin->output((string)"X-Envelope-To: "+assassin->rcpt()+"\r\n");
-		assassin->output((string)"Received: from "+macro_s+" ("+smfi_getsymval(ctx,"_")+") by "+smfi_getsymval(ctx,"j")+"; "+macro_b+"\r\n");
+		assassin->output((string)"X-Envelope-To: "+envrcpt[0]+"\r\n");
+
+		if (!macro_b)
+			assassin->output((string)"Received: from "+macro_s+" ("+macro__+") by "+macro_j+";\r\n");
+		else
+			assassin->output((string)"Received: from "+macro_s+" ("+macro__+") by "+macro_j+"; "+macro_b+"\r\n");
 
 	} else
+		assassin->output((string)"X-Envelope-To: "+envrcpt[0]+"\r\n");
+
+	/* increment RCPT TO: count */
+	assassin->set_numrcpt();
+
+	/* If we expanded to at least one user and we haven't recorded one yet,
+	   record the first one */
+	if (!assassin->expandedrcpt.empty() && (assassin->rcpt().size() == 0))
 	{
-		assassin->set_numrcpt();
-	}
-	for( rcpt = envrcpt; *rcpt; rcpt++ ) 
-	{
-		assassin->recipients.push_back( *rcpt ); // XXX verify that this worked
+		debug(D_RCPT, "remembering %s for spamc", assassin->expandedrcpt.front().c_str());
+		assassin->set_rcpt(assassin->expandedrcpt.front());
 	}
+
+	debug(D_RCPT, "remembering recipient %s", envrcpt[0]);
+	assassin->recipients.push_back( envrcpt[0] ); // XXX verify that this worked
+
+	debug(D_FUNC, "mlfi_envrcpt: exit");
+
 	return SMFIS_CONTINUE;
 }
 
@@ -749,28 +985,31 @@ mlfi_header(SMFICTX* ctx, char* headerf, char* headerv)
        } 
        catch (string& problem) {
          throw_error(problem);
+         ((struct context *)smfi_getpriv(ctx))->assassin=NULL;
+         delete assassin;
+         debug(D_FUNC, "mlfi_header: exit error connect");
          return SMFIS_TEMPFAIL;
        };
      }
 
   // Is it a "X-Spam-" header field?
-  if ( cmp_nocase_partial(string("X-Spam-"), string(headerf)) == 0 )
+  if ( cmp_nocase_partial("X-Spam-", headerf) == 0 )
     {
       int suppress = 1;
       // memorize content of old fields
 
-      if ( cmp_nocase_partial(string("X-Spam-Status"), string(headerf)) == 0 )
-	assassin->set_spam_status(string(headerv));
-      else if ( cmp_nocase_partial(string("X-Spam-Flag"), string(headerf)) == 0 )
-	assassin->set_spam_flag(string(headerv));
-      else if ( cmp_nocase_partial(string("X-Spam-Report"), string(headerf)) == 0 )
-	assassin->set_spam_report(string(headerv));
-      else if ( cmp_nocase_partial(string("X-Spam-Prev-Content-Type"), string(headerf)) == 0 )
-	assassin->set_spam_prev_content_type(string(headerv));
-      else if ( cmp_nocase_partial(string("X-Spam-Level"), string(headerf)) == 0 )
-	assassin->set_spam_level(string(headerv));
-      else if ( cmp_nocase_partial(string("X-Spam-Checker-Version"), string(headerf)) == 0 )
-	assassin->set_spam_checker_version(string(headerv));
+      if ( cmp_nocase_partial("X-Spam-Status", headerf) == 0 )
+	assassin->set_spam_status(headerv);
+      else if ( cmp_nocase_partial("X-Spam-Flag", headerf) == 0 )
+	assassin->set_spam_flag(headerv);
+      else if ( cmp_nocase_partial("X-Spam-Report", headerf) == 0 )
+	assassin->set_spam_report(headerv);
+      else if ( cmp_nocase_partial("X-Spam-Prev-Content-Type", headerf) == 0 )
+	assassin->set_spam_prev_content_type(headerv);
+      else if ( cmp_nocase_partial("X-Spam-Level", headerf) == 0 )
+	assassin->set_spam_level(headerv);
+      else if ( cmp_nocase_partial("X-Spam-Checker-Version", headerf) == 0 )
+	assassin->set_spam_checker_version(headerv);
       else
       {
       	/* Hm. X-Spam header, but not one we recognize.  Pass it through. */
@@ -782,19 +1021,18 @@ mlfi_header(SMFICTX* ctx, char* headerf, char* headerv)
 	debug(D_FUNC, "mlfi_header: suppress");
 	return SMFIS_CONTINUE;
       }
-    };
+    }
 
   // Content-Type: will be stored if present
-  if ( cmp_nocase_partial(string("Content-Type"), string(headerf)) == 0 )
-    assassin->set_content_type(string(headerv));
+  if ( cmp_nocase_partial("Content-Type", headerf) == 0 )
+    assassin->set_content_type(headerv);
 
   // Subject: should be stored
-  if ( cmp_nocase_partial(string("Subject"), string(headerf)) == 0 )
-    assassin->set_subject(string(headerv));
+  if ( cmp_nocase_partial("Subject", headerf) == 0 )
+    assassin->set_subject(headerv);
 
   // assemble header to be written to SpamAssassin
-  string header=string(headerf)+string(": ")+
-    string(headerv)+string("\r\n");
+  string header = string(headerf) + ": " + headerv + "\r\n";
  
   try {
     // write to SpamAssassin client
@@ -804,7 +1042,7 @@ mlfi_header(SMFICTX* ctx, char* headerf, char* headerv)
       throw_error(problem);
       ((struct context *)smfi_getpriv(ctx))->assassin=NULL;
       delete assassin;
-      debug(D_FUNC, "mlfi_header: exit error");
+      debug(D_FUNC, "mlfi_header: exit error output");
       return SMFIS_TEMPFAIL;
     };
   
@@ -836,6 +1074,10 @@ mlfi_eoh(SMFICTX* ctx)
        } 
        catch (string& problem) {
          throw_error(problem);
+         ((struct context *)smfi_getpriv(ctx))->assassin=NULL;
+         delete assassin;
+
+         debug(D_FUNC, "mlfi_eoh: exit error connect");
          return SMFIS_TEMPFAIL;
        };
      }
@@ -849,7 +1091,7 @@ mlfi_eoh(SMFICTX* ctx)
       ((struct context *)smfi_getpriv(ctx))->assassin=NULL;
       delete assassin;
   
-      debug(D_FUNC, "mlfi_eoh: exit error");
+      debug(D_FUNC, "mlfi_eoh: exit error output");
       return SMFIS_TEMPFAIL;
     };
   
@@ -982,6 +1224,7 @@ mlfi_abort(SMFICTX* ctx)
 //
 SpamAssassin::SpamAssassin():
   error(false),
+  running(false),
   connected(false),
   _numrcpt(0)
 {
@@ -1001,12 +1244,15 @@ SpamAssassin::~SpamAssassin()
 		// child still running?
 		if (running)
 		{
-			// slaughter child
-			kill(pid, SIGKILL);
+			// make sure the pid is valid
+			if (pid > 0) {
+				// slaughter child
+				kill(pid, SIGKILL);
 
-			// wait for child to terminate
-			int status;
-			waitpid(pid, &status, 0);
+				// wait for child to terminate
+				int status;
+				waitpid(pid, &status, 0);
+			}
 		}
     }
 
@@ -1015,6 +1261,11 @@ SpamAssassin::~SpamAssassin()
 	{
 		recipients.pop_front();
 	}
+	// Clean up the recip list. Might be overkill, but it's good housekeeping.
+	while( !expandedrcpt.empty()) 
+	{
+		expandedrcpt.pop_front();
+	}
 }
 
 //
@@ -1029,7 +1280,7 @@ void SpamAssassin::Connect()
     throw string(string("pipe error: ")+string(strerror(errno)));
   if (pipe(pipe_io[1]))
     throw string(string("pipe error: ")+string(strerror(errno)));
-  
+
   // now execute SpamAssassin client for contact with SpamAssassin spamd
 
   // start child process
@@ -1063,19 +1314,25 @@ void SpamAssassin::Connect()
       if (flag_sniffuser) 
       {
         argv[argc++] = "-u";
-        if ( numrcpt() != 1 )
+        if ( expandedrcpt.size() != 1 )
         {
           // More (or less?) than one recipient, so we pass the default
           // username to SPAMC.  This way special rules can be defined for
           // multi recipient messages.
+          debug(D_RCPT, "%d recipients; spamc gets default username %s", (int)expandedrcpt.size(), defaultuser);
           argv[argc++] = defaultuser; 
         } else
         { 
           // There is only 1 recipient so we pass the username
           // (converted to lowercase) to SPAMC.  Don't worry about 
           // freeing this memory as we're exec()ing anyhow.
-          argv[argc] = strlwr(strdup(local_user().c_str())); 
-          
+          if (flag_full_email)
+            argv[argc] = strlwr(strdup(full_user().c_str())); 
+          else
+            argv[argc] = strlwr(strdup(local_user().c_str())); 
+
+          debug(D_RCPT, "spamc gets %s", argv[argc]);
+         
           argc++;
         }
       }
@@ -1095,7 +1352,7 @@ void SpamAssassin::Connect()
 
       // execution failed
       throw_error(string("execution error: ")+string(strerror(errno)));
-      
+      _exit(1);
       break;
     }
 
@@ -1146,11 +1403,13 @@ SpamAssassin::output(const void* buffer, long size)
 		(string::size_type, char), so we have to cast the parameters.
 	*/
   	outputbuffer.append((const char *)buffer,(string::size_type)size);
+  	debug(D_FUNC, "::output exit1");
   	return;
   }
 
   // send to SpamAssassin
-  long total(0), wsize(0);
+  long total = 0;
+  long wsize = 0;
   string reason;
   int status;
   do {
@@ -1168,6 +1427,16 @@ SpamAssassin::output(const void* buffer, long size)
 
 	debug(D_POLL, "poll returned %d, fd0=%d, fd1=%d", nready, fds[0].revents, fds[1].revents);
 
+	if (fds[1].revents & (POLLERR|POLLNVAL|POLLHUP))
+	{
+		throw string("poll says my read pipe is busted");
+	}
+
+	if (fds[0].revents & (POLLERR|POLLNVAL|POLLHUP))
+	{
+		throw string("poll says my write pipe is busted");
+	}
+
 	if (fds[1].revents & POLLIN)
 	{
 		debug(D_POLL, "poll says I can read");
@@ -1210,7 +1479,7 @@ SpamAssassin::output(const void* buffer, long size)
 	}
   } while ( total < size );
 
-  debug(D_FUNC, "::output exit");
+  debug(D_FUNC, "::output exit2");
 }
 
 void SpamAssassin::output(const void* buffer)
@@ -1235,11 +1504,14 @@ SpamAssassin::close_output()
 void
 SpamAssassin::input()
 {
-	debug(D_FUNC, "::input enter");
+  debug(D_FUNC, "::input enter");
   // if the child has exited or we experienced an error, return
   // immediately.
   if (!running || error)
+  {
+    debug(D_FUNC, "::input exit1");
     return;
+  }
 
   // keep reading from input pipe until it is empty
   empty_and_close_pipe();
@@ -1254,7 +1526,7 @@ SpamAssassin::input()
       error = true;
       throw string(string("waitpid error: ")+string(strerror(errno)));
     }; 
-	debug(D_FUNC, "::input exit");
+	debug(D_FUNC, "::input exit2");
 }
 
 //
@@ -1340,8 +1612,29 @@ string
 SpamAssassin::local_user()
 {
   // assuming we have a recipient in the form: <username@somehost.somedomain>
-  // we return 'username'
-  return _rcpt.substr(1,_rcpt.find('@')-1);
+  // (angle brackets optional) we return 'username'
+  if (_rcpt[0] == '<')
+    return _rcpt.substr(1, _rcpt.find_first_of("@+")-1);
+  else
+  	return _rcpt.substr(0, _rcpt.find_first_of("@+"));
+}
+
+string
+SpamAssassin::full_user()
+{
+  string name;
+  // assuming we have a recipient in the form: <username@somehost.somedomain>
+  // (angle brackets optional) we return 'username@somehost.somedomain'
+  if (_rcpt[0] == '<')
+    name = _rcpt.substr(1, _rcpt.find('>')-1);
+  else
+  	name = _rcpt;
+  if(name.find('@') == string::npos)
+  {
+    /* if the name had no domain part (local delivery), append the default one */
+    name = name + "@" + defaultdomain;
+  }
+  return name;
 }
 
 int
@@ -1539,7 +1832,7 @@ void
 throw_error(const string& errmsg)
 {
   if (errmsg.c_str())
-    syslog(LOG_ERR, errmsg.c_str());
+    syslog(LOG_ERR, "Thrown error: %s", errmsg.c_str());
   else
     syslog(LOG_ERR, "Unknown error");
 }
@@ -1551,6 +1844,9 @@ void parse_debuglevel(char* string)
 {
 	char *token;
 
+	/* make a copy so we don't overwrite argv[] */
+	string = strdup(string);
+
 	/* handle the old numeric values too */
 	switch(atoi(string))
 	{
@@ -1561,6 +1857,7 @@ void parse_debuglevel(char* string)
 		case 1:
 			flag_debug |= (1<<D_MISC) | (1<<D_FUNC);
 			debug(D_ALWAYS, "Setting debug level to 0x%0x", flag_debug);
+			free(string);
 			return;
 		default:
 			break;
@@ -1590,6 +1887,7 @@ void parse_debuglevel(char* string)
 		}
 	}
 	debug(D_ALWAYS, "Setting debug level to 0x%0x", flag_debug);
+	free(string);
 }
 
 /*
@@ -1640,7 +1938,7 @@ find_nocase(const string& array, const string& pattern, string::size_type start)
 
   while (pos < array.size())
     {
-      string::size_type ctr(0);
+      string::size_type ctr = 0;
 
       while( (pos+ctr) < array.size() &&
 	     toupper(array[pos+ctr]) == toupper(pattern[ctr]) )
@@ -1694,6 +1992,9 @@ void parse_networklist(char *string, struct networklist *list)
 {
 	char *token;
 
+	/* make a copy so we don't overwrite argv[] */
+	string = strdup(string);
+
 	while ((token = strsep(&string, ", ")))
 	{
 		char *tnet = strsep(&token, "/");
@@ -1742,12 +2043,16 @@ void parse_networklist(char *string, struct networklist *list)
 		list->nets[list->num_nets].netmask = mask;
 		list->num_nets++;
 	}
+	free(string);
 }
 
 int ip_in_networklist(struct in_addr ip, struct networklist *list)
 {
 	int i;
 
+	if (list->num_nets == 0)
+		return 0;
+		
 	debug(D_NET, "Checking %s against:", inet_ntoa(ip));
 	for (i = 0; i < list->num_nets; i++)
 	{
