@@ -1,6 +1,6 @@
 //-*-c++-*-
 //
-//  $Id: spamass-milter.h,v 1.17 2003/06/14 19:17:41 dnelson Exp $
+//  $Id: spamass-milter.h,v 1.22 2004/09/21 20:51:06 dnelson Exp $
 //
 //  Main include file for SpamAss-Milter
 //
@@ -25,6 +25,16 @@
 //   Contact:
 //            Michael Brown <michaelb@opentext.com>
 //
+#ifndef _SPAMASS_MILTER_H
+#define _SPAMASS_MILTER_H
+
+#ifdef HAVE_CDEFS_H
+#include <sys/cdefs.h>
+#endif
+#if !defined(__printflike)
+#define __printflike(a,b)
+#endif
+
 #include <list>
 
 using namespace std;
@@ -63,7 +73,8 @@ struct networklist
 // Debug tokens.
 enum debuglevel 
 {
-	D_ALWAYS, D_FUNC, D_POLL, D_UORI, D_STR, D_MISC, D_NET, D_SPAMC,
+	D_ALWAYS, D_FUNC, D_POLL, D_UORI, D_STR, D_MISC, D_NET, D_SPAMC, D_RCPT,
+	D_COPY,
 	D_MAX // must be last
 };
 
@@ -89,13 +100,14 @@ public:
   string& spam_level();
   string& content_type();
   string& subject();
-  string& rcpt();
-  string& from();
-  string& connectip();
-  string  local_user();
-  int     numrcpt();
-  int     set_numrcpt();
-  int     set_numrcpt(const int);
+  string& rcpt();		/* first RCPT TO: recipient (raw) */
+  string& from();		/* MAIL FROM: sender (raw) */
+  string& connectip();	/* IP of sending machine */
+  string  local_user();	/* username part of first expanded recipient */
+  string  full_user();	/* full first expanded recipient */
+  int     numrcpt();	/* total RCPT TO: recpients */
+  int     set_numrcpt();	/* increment total RCPT count */
+  int     set_numrcpt(const int);	/* set total RCPT count to n */
   string::size_type set_spam_status(const string&);
   string::size_type set_spam_flag(const string&);
   string::size_type set_spam_report(const string&);
@@ -115,7 +127,7 @@ private:
 public:  
   // flags
   bool error;
-  bool running;
+  bool running;		/* XXX merge running, connected, and pid */
   bool connected;	/* are we connected to spamc? */
 
   // This is where we store the mail after it
@@ -139,6 +151,9 @@ public:
   // The list of recipients for the current message
   list <string> recipients;
 
+  // List of recipients after alias/virtusertable expansion
+  list <string> expandedrcpt;
+
   // the sendmail queue id for this message; used for logging
   string queueid;
 
@@ -162,7 +177,7 @@ typedef string::size_type (SpamAssassin::*t_setter)(const string &val);
 int assassinate(SMFICTX*, SpamAssassin*);
 
 void throw_error(const string&);
-void debug(enum debuglevel, const char* fmt, ...); /* __printflike(2, 3); */
+void debug(enum debuglevel, const char* fmt, ...) __printflike(2, 3);
 string::size_type find_nocase(const string&, const string&, string::size_type = 0);
 int cmp_nocase_partial(const string&, const string&);
 void closeall(int fd);
@@ -170,3 +185,5 @@ void parse_networklist(char *string, struct networklist *list);
 int ip_in_networklist(struct in_addr ip, struct networklist *list);
 void parse_debuglevel(char* string);
 char *strlwr(char *str);
+
+#endif
